﻿using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Configuration.Provider;
using System.Linq;
using System.Text;

using System.Web;
using System.Web.Security;
using System.Web.Configuration;
using System.Web.Profile;
using System.Web.Caching;

using System.Security;
using System.Security.Principal;

using Norm;
using Norm.Attributes;
using Norm.Linq;


namespace EazyProviders.MongoDb
{


    public class MongoRoleProvider : RoleProvider
    {
        private string _AppName;
        private int? _AppId;
        private string _mongoConnectionString;

        

        public override void Initialize(string name, NameValueCollection config)
        {
            if (config == null)
                throw new ArgumentNullException("config");

            if (String.IsNullOrEmpty(name))
                name = "EazyRoleProviderMongoDb";
            if (string.IsNullOrEmpty(config["description"]))
            {
                config.Remove("description");
                config.Add("description", "EazyRole Provider for MongoDb");
            }
            base.Initialize(name, config);


            string temp = config["connectionStringName"];
            
            if (string.IsNullOrEmpty(temp))
                throw new ProviderException(StringResources.GetString(StringResources.Connection_name_not_specified));
            
            _mongoConnectionString = SecUtility.GetConnectionString(temp, true, true);

            if (string.IsNullOrEmpty(_mongoConnectionString))
            {
                throw new ProviderException(StringResources.GetString(StringResources.Connection_string_not_found, temp));
            }

            _AppName = config["applicationName"];

            if (string.IsNullOrEmpty(_AppName))
                _AppName = SecUtility.GetDefaultAppName();

            if (_AppName.Length > 256)
            {
                throw new ProviderException(StringResources.GetString(StringResources.Provider_application_name_too_long));
            }

            config.Remove("connectionStringName");
            config.Remove("applicationName");
            config.Remove("commandTimeout");

            if (config.Count > 0)
            {
                string attribUnrecognized = config.GetKey(0);
                if (!String.IsNullOrEmpty(attribUnrecognized))
                    throw new ProviderException(StringResources.GetString(StringResources.Provider_unrecognized_attribute, attribUnrecognized));
            }

            InitMembershipCollections();

        }


        void InitMembershipCollections()
        {
            using (var session = new MongoSession(_mongoConnectionString))
            {
                var app = (from a in session.Applications
                           where a.ApplicationName == ApplicationName
                           select a).SingleOrDefault();

                if (app == null)
                {
                    app = new MembershipApplication
                                  {
                                      ApplicationName = _AppName,
                                      LoweredApplicationName = _AppName.ToLowerInvariant()
                                  };

                    session.Add(app);
                }

                _AppId = app.ApplicationId;

            }
        }
       
        
        public override bool IsUserInRole(string username, string roleName)
        {
            SecUtility.CheckParameter(ref roleName, true, true, true, 256, "roleName");
            SecUtility.CheckParameter(ref username, true, false, true, 256, "username");
            if (username.Length < 1)
                return false;

            try
            {

                using (var session = new MongoSession(_mongoConnectionString))
                {

                    //var check = (from r in session.Roles
                    //             where
                    //                 r.Users.Any(u => u.UserName == username)
                    //                 && r.RoleName == roleName
                    //                 && r.ApplicationId==_AppId select r).Any();



                    var user = (from u in session.Users
                                 where u.UserName == username
                                     && u.ApplicationId == _AppId select u).SingleOrDefault();

                    if (user == null)
                        return false;

                    if(user.Roles==null)
                        return false;

                    if(user.Roles.Where(r=>r.RoleName==roleName).Any())
                        return true;

                    return false;

                }


            }
            catch
            {
                throw;
            }

        }

       
        public override string[] GetRolesForUser(string username)
        {
            SecUtility.CheckParameter(ref username, true, false, true, 256, "username");
            if (username.Length < 1)
                return new string[0];

            try
            {

                using (var session = new MongoSession(_mongoConnectionString))
                {

                    var user = (from u in session.Users
                                where u.UserName == username
                                    && u.ApplicationId == _AppId
                                select u).SingleOrDefault();

                    if (user == null)
                        return new string[0];

                    if (user.Roles == null)
                        return new string[0];

                    return user.Roles.Select(r => r.RoleName).ToArray();
                    

                }


            }
            catch
            {
                throw;
            }
        }

      
        public override void CreateRole(string roleName)
        {
            SecUtility.CheckParameter(ref roleName, true, true, true, 256, "roleName");
            try
            {

                using (var session = new MongoSession(_mongoConnectionString))
                {

                    var roles = from r in session.Roles
                                where
                                    r.RoleName==roleName
                                    && r.ApplicationId == _AppId
                                select r;

                    if (roles.Any())
                    {
                        throw new ProviderException(StringResources.GetString(StringResources.Provider_role_already_exists, roleName));
                    }

                    
                    var role = new MembershipRole
                                   {
                                       ApplicationId=_AppId,
                                       RoleName = roleName, 
                                       LoweredRoleName = roleName.ToLowerInvariant()
                                   };

                    session.Add(role);
                }

            }
            catch
            {
                throw;
            }
        }

        
        public override bool DeleteRole(string roleName, bool throwOnPopulatedRole)
        {
            SecUtility.CheckParameter(ref roleName, true, true, true, 256, "roleName");
            try
            {

                using (var session = new MongoSession(_mongoConnectionString))
                {

                    var role = (from r in session.Roles
                                where
                                    r.RoleName == roleName
                                    && r.ApplicationId == _AppId
                                select r).SingleOrDefault();


                    var users = (from u in session.Users
                                 where
                                     u.ApplicationId == _AppId &&
                                     u.Roles != null &&
                                     u.Roles.Where(r => r.RoleName == roleName).Any()
                                 select u);

                    if (users.Any() && throwOnPopulatedRole)
                    {
                        throw new ProviderException(StringResources.GetString(StringResources.Role_is_not_empty));
                    }

                    session.Delete(role);

                    return true;

                }
                
               
            }
            catch
            {
                throw;
            }
        }

        
        public override bool RoleExists(string roleName)
        {
            SecUtility.CheckParameter(ref roleName, true, true, true, 256, "roleName");

            try
            {
                using (var session = new MongoSession(_mongoConnectionString))
                {

                    var role = (from r in session.Roles
                                where
                                    r.RoleName == roleName
                                    && r.ApplicationId == _AppId
                                select r).SingleOrDefault();


                    return (role != null);

                }
            }
            catch
            {
                throw;
            }
        }

       

        public override void AddUsersToRoles(string[] usernames, string[] roleNames)
        {
            SecUtility.CheckArrayParameter(ref roleNames, true, true, true, 256, "roleNames");
            SecUtility.CheckArrayParameter(ref usernames, true, true, true, 256, "usernames");

            try
            {

                List<string> _usernames = usernames.ToList();
                List<string> _roleNames = roleNames.ToList();

                using (var session = new MongoSession(_mongoConnectionString))
                {

                    var users = (from u in session.Users
                                where u.ApplicationId == _AppId && _usernames.Contains(u.UserName)
                                select u).ToList();

                    var roles = (from r in session.Roles
                                where r.ApplicationId == _AppId && _roleNames.Contains(r.RoleName)
                                select r).ToList();

                    foreach (var userEntity in users)
                    {
                        if (userEntity.Roles != null && userEntity.Roles.Any())
                        {
                            var newRoles = roles.Except(userEntity.Roles);
                            userEntity.Roles.AddRange(newRoles);

                        }
                        else
                        {
                            userEntity.Roles=new List<MembershipRole>();
                            userEntity.Roles.AddRange(roles);
                        }

                        session.Save(userEntity);
                        
                    }


                }
 
            }
            catch
            {
                throw;
            }
        }


        public override void RemoveUsersFromRoles(string[] usernames, string[] roleNames)
        {
            SecUtility.CheckArrayParameter(ref roleNames, true, true, true, 256, "roleNames");
            SecUtility.CheckArrayParameter(ref usernames, true, true, true, 256, "usernames");

            try
            {
                List<string> _usernames = usernames.ToList();
                List<string> _roleNames = roleNames.ToList();

                using (var session = new MongoSession(_mongoConnectionString))
                {

                    var users = (from u in session.Users
                                where u.ApplicationId == _AppId && _usernames.Contains(u.UserName)
                                select u).ToList();

                    var roles = (from r in session.Roles
                                where r.ApplicationId == _AppId && _roleNames.Contains(r.RoleName)
                                select r).ToList();

                    foreach (var userEntity in users)
                    {
                        if (userEntity.Roles != null && userEntity.Roles.Any())
                        {
                            var matchedRoles = roles.Intersect(userEntity.Roles);

                            foreach (var matchedRole in matchedRoles)
                                userEntity.Roles.Remove(matchedRole);

                            if (matchedRoles.Any())
                                session.Save(userEntity);

                        }

                        

                    }


                }

            }
            catch
            {
                throw;
            }
        }



        public override string[] GetUsersInRole(string roleName)
        {
            SecUtility.CheckParameter(ref roleName, true, true, true, 256, "roleName");

            try
            {

                using (var session = new MongoSession(_mongoConnectionString))
                {

                    var role = (from r in session.Roles
                                    where r.ApplicationId == _AppId &&
                                          r.RoleName == roleName
                                    select r).SingleOrDefault();

                    if (role == null)
                    {
                        throw new ProviderException(StringResources.GetString(StringResources.Provider_role_not_found, roleName));
                    }

                    var users = from u in session.Users
                                where u.ApplicationId == _AppId &&
                                u.Roles.Any(r => r.RoleName == roleName)
                                select u;

                    if (users==null || !users.Any())
                        return new string[0];

                    return users.Select(u => u.UserName).ToArray();
                    
                }

            }
            catch
            {
                throw;
            }
        }

       

        public override string[] GetAllRoles()
        {
            try
            {
                using (var session = new MongoSession(_mongoConnectionString))
                {

                    var roles = from r in session.Roles
                                where r.ApplicationId == _AppId
                                select r;

                    if (roles == null || !roles.Any())
                        return new string[0];

                    return roles.Select(u => u.RoleName).ToArray();

                }
            }
            catch
            {
                throw;
            }
        }

      
        public override string[] FindUsersInRole(string roleName, string usernameToMatch)
        {
            SecUtility.CheckParameter(ref roleName, true, true, true, 256, "roleName");
            SecUtility.CheckParameter(ref usernameToMatch, true, true, false, 256, "usernameToMatch");

            try
            {

                using (var session = new MongoSession(_mongoConnectionString))
                {

                    var role = (from r in session.Roles
                                where r.ApplicationId == _AppId &&
                                      r.RoleName == roleName
                                select r).SingleOrDefault();

                    if (role == null)
                    {
                        throw new ProviderException(StringResources.GetString(StringResources.Provider_role_not_found, roleName));
                    }

                    var users = from u in session.Users
                                where u.ApplicationId == _AppId &&
                                u.UserName == usernameToMatch &&
                                u.Roles.Any(r => r.RoleName == roleName)
                                select u;

                    if (users == null || !users.Any())
                        return new string[0];

                    return users.Select(u => u.UserName).ToArray();

                }

            }
            catch
            {
                throw;
            }
        }
       
        
        
        public override string ApplicationName
        {
            get { return _AppName; }
            set
            {
                _AppName = value;

                if (_AppName.Length > 256)
                {
                    throw new ProviderException(StringResources.GetString(StringResources.Provider_application_name_too_long));
                }
            }
        }

       
    }

}
